@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, see
- * <http://www.gnu.org/licenses/>.
+ * <https://www.gnu.org/licenses/>.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -67,15 +67,10 @@ static void set_location_internal (GWeatherLocationEntry *entry,
 				   GtkTreeModel          *model,
 				   GtkTreeIter           *iter,
 				   GWeatherLocation      *loc);
-static GWeatherLocation *
-create_new_detached_location (GWeatherLocation *nearest_station,
-                              const char       *name,
-                              gboolean          latlon_valid,
-                              gdouble           latitude,
-                              gdouble           longitude);
 static void
-fill_location_entry_model (GtkTreeStore *store, GWeatherLocation *loc,
+fill_location_entry_model (GtkListStore *store, GWeatherLocation *loc,
 			   const char *parent_display_name,
+			   const char *parent_sort_local_name,
 			   const char *parent_compare_local_name,
 			   const char *parent_compare_english_name,
 			   gboolean show_named_timezones);
@@ -84,6 +79,7 @@ enum LOC
 {
     LOC_GWEATHER_LOCATION_ENTRY_COL_DISPLAY_NAME = 0,
     LOC_GWEATHER_LOCATION_ENTRY_COL_LOCATION,
+    LOC_GWEATHER_LOCATION_ENTRY_COL_LOCAL_SORT_NAME,
     LOC_GWEATHER_LOCATION_ENTRY_COL_LOCAL_COMPARE_NAME,
     LOC_GWEATHER_LOCATION_ENTRY_COL_ENGLISH_COMPARE_NAME,
     LOC_GWEATHER_LOCATION_ENTRY_NUM_COLUMNS
@@ -93,6 +89,7 @@ enum PLACE
 {
     PLACE_GWEATHER_LOCATION_ENTRY_COL_DISPLAY_NAME = 0,
     PLACE_GWEATHER_LOCATION_ENTRY_COL_PLACE,
+    PLACE_GWEATHER_LOCATION_ENTRY_COL_LOCAL_SORT_NAME,
     PLACE_GWEATHER_LOCATION_ENTRY_COL_LOCAL_COMPARE_NAME
 };
 
@@ -171,20 +168,41 @@ dispose (GObject *object)
     G_OBJECT_CLASS (gweather_location_entry_parent_class)->dispose (object);
 }
 
+static int
+tree_compare_local_name (GtkTreeModel *model,
+			 GtkTreeIter  *a,
+			 GtkTreeIter  *b,
+			 gpointer      user_data)
+{
+    g_autofree gchar *name_a = NULL, *name_b = NULL;
+
+    gtk_tree_model_get (model, a,
+			LOC_GWEATHER_LOCATION_ENTRY_COL_LOCAL_SORT_NAME, &name_a,
+			-1);
+    gtk_tree_model_get (model, b,
+			LOC_GWEATHER_LOCATION_ENTRY_COL_LOCAL_SORT_NAME, &name_b,
+			-1);
+
+    return g_utf8_collate (name_a, name_b);
+}
+
+
 static void
 constructed (GObject *object)
 {
     GWeatherLocationEntry *entry;
-    GtkTreeStore *store = NULL;
+    GtkListStore *store = NULL;
     GtkEntryCompletion *completion;
 
     entry = GWEATHER_LOCATION_ENTRY (object);
 
     if (!entry->priv->top)
-	entry->priv->top = gweather_location_ref (gweather_location_get_world ());
+	entry->priv->top = gweather_location_get_world ();
 
-    store = gtk_tree_store_new (4, G_TYPE_STRING, GWEATHER_TYPE_LOCATION, G_TYPE_STRING, G_TYPE_STRING);
-    fill_location_entry_model (store, entry->priv->top, NULL, NULL, NULL, entry->priv->show_named_timezones);
+    store = gtk_list_store_new (5, G_TYPE_STRING, GWEATHER_TYPE_LOCATION, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
+    gtk_tree_sortable_set_default_sort_func (GTK_TREE_SORTABLE (store),
+                                             tree_compare_local_name, NULL, NULL);
+    fill_location_entry_model (store, entry->priv->top, NULL, NULL, NULL, NULL, entry->priv->show_named_timezones);
 
     entry->priv->model = GTK_TREE_MODEL (store);
     completion = gtk_entry_get_completion (GTK_ENTRY (entry));
@@ -325,7 +343,7 @@ set_location_internal (GWeatherLocationEntry *entry,
 	g_free (name);
     } else if (loc) {
 	priv->location = gweather_location_ref (loc);
-	gtk_entry_set_text (GTK_ENTRY (entry), loc->local_name);
+	gtk_entry_set_text (GTK_ENTRY (entry), gweather_location_get_name (loc));
 	priv->custom_text = FALSE;
     } else {
 	priv->location = NULL;
@@ -489,56 +507,57 @@ gweather_location_entry_set_city (GWeatherLocationEntry *entry,
 }
 
 static void
-fill_location_entry_model (GtkTreeStore *store, GWeatherLocation *loc,
+fill_location_entry_model (GtkListStore *store, GWeatherLocation *loc,
 			   const char *parent_display_name,
+			   const char *parent_sort_local_name,
 			   const char *parent_compare_local_name,
 			   const char *parent_compare_english_name,
 			   gboolean show_named_timezones)
 {
-    GWeatherLocation **children;
-    char *display_name, *local_compare_name, *english_compare_name;
-    int i;
-
-    children = gweather_location_get_children (loc);
+    g_autoptr(GWeatherLocation) child = NULL;
+    char *display_name, *local_sort_name, *local_compare_name, *english_compare_name;
 
     switch (gweather_location_get_level (loc)) {
     case GWEATHER_LOCATION_WORLD:
     case GWEATHER_LOCATION_REGION:
-    case GWEATHER_LOCATION_ADM2:
 	/* Ignore these levels of hierarchy; just recurse, passing on
 	 * the names from the parent node.
 	 */
-	for (i = 0; children[i]; i++) {
-	    fill_location_entry_model (store, children[i],
+	while ((child = gweather_location_next_child (loc, child)))
+	    fill_location_entry_model (store, child,
 				       parent_display_name,
+				       parent_sort_local_name,
 				       parent_compare_local_name,
 				       parent_compare_english_name,
 				       show_named_timezones);
-	}
 	break;
 
     case GWEATHER_LOCATION_COUNTRY:
 	/* Recurse, initializing the names to the country name */
-	for (i = 0; children[i]; i++) {
-	    fill_location_entry_model (store, children[i],
-				       loc->local_name,
-				       loc->local_sort_name,
-				       loc->english_sort_name,
+	while ((child = gweather_location_next_child (loc, child)))
+	    fill_location_entry_model (store, child,
+				       gweather_location_get_name (loc),
+				       gweather_location_get_sort_name (loc),
+				       gweather_location_get_sort_name (loc),
+				       gweather_location_get_english_sort_name (loc),
 				       show_named_timezones);
-	}
 	break;
 
     case GWEATHER_LOCATION_ADM1:
 	/* Recurse, adding the ADM1 name to the country name */
-	display_name = g_strdup_printf ("%s, %s", loc->local_name, parent_display_name);
-	local_compare_name = g_strdup_printf ("%s, %s", loc->local_sort_name, parent_compare_local_name);
-	english_compare_name = g_strdup_printf ("%s, %s", loc->english_sort_name, parent_compare_english_name);
+	/* Translators: this is the name of a location followed by a region, for example:
+	 * 'London, United Kingdom'
+	 * You shouldn't need to translate this string unless the language has a different comma.
+	 */
+	display_name = g_strdup_printf (_("%s, %s"), gweather_location_get_name (loc), parent_display_name);
+	local_sort_name = g_strdup_printf ("%s, %s", parent_sort_local_name, gweather_location_get_sort_name (loc));
+	local_compare_name = g_strdup_printf ("%s, %s", gweather_location_get_sort_name (loc), parent_compare_local_name);
+	english_compare_name = g_strdup_printf ("%s, %s", gweather_location_get_english_sort_name (loc), parent_compare_english_name);
 
-	for (i = 0; children[i]; i++) {
-	    fill_location_entry_model (store, children[i],
-				       display_name, local_compare_name, english_compare_name,
+	while ((child = gweather_location_next_child (loc, child)))
+	    fill_location_entry_model (store, child,
+				       display_name, local_sort_name, local_compare_name, english_compare_name,
 				       show_named_timezones);
-	}
 
 	g_free (display_name);
 	g_free (local_compare_name);
@@ -554,16 +573,23 @@ fill_location_entry_model (GtkTreeStore *store, GWeatherLocation *loc,
 	 */
     case GWEATHER_LOCATION_WEATHER_STATION:
 	/* <location> with no parent <city> */
-	display_name = g_strdup_printf ("%s, %s",
-					loc->local_name, parent_display_name);
+	/* Translators: this is the name of a location followed by a region, for example:
+	 * 'London, United Kingdom'
+	 * You shouldn't need to translate this string unless the language has a different comma.
+	 */
+	display_name = g_strdup_printf (_("%s, %s"),
+					gweather_location_get_name (loc), parent_display_name);
+	local_sort_name = g_strdup_printf ("%s, %s",
+					   parent_sort_local_name, gweather_location_get_sort_name (loc));
 	local_compare_name = g_strdup_printf ("%s, %s",
-					      loc->local_sort_name, parent_compare_local_name);
+					      gweather_location_get_sort_name (loc), parent_compare_local_name);
 	english_compare_name = g_strdup_printf ("%s, %s",
-						loc->english_sort_name, parent_compare_english_name);
+						gweather_location_get_english_sort_name (loc), parent_compare_english_name);
 
-	gtk_tree_store_insert_with_values (store, NULL, NULL, -1,
+	gtk_list_store_insert_with_values (store, NULL, -1,
 					   LOC_GWEATHER_LOCATION_ENTRY_COL_LOCATION, loc,
 					   LOC_GWEATHER_LOCATION_ENTRY_COL_DISPLAY_NAME, display_name,
+					   LOC_GWEATHER_LOCATION_ENTRY_COL_LOCAL_SORT_NAME, local_sort_name,
 					   LOC_GWEATHER_LOCATION_ENTRY_COL_LOCAL_COMPARE_NAME, local_compare_name,
 					   LOC_GWEATHER_LOCATION_ENTRY_COL_ENGLISH_COMPARE_NAME, english_compare_name,
 					   -1);
@@ -575,11 +601,12 @@ fill_location_entry_model (GtkTreeStore *store, GWeatherLocation *loc,
 
     case GWEATHER_LOCATION_NAMED_TIMEZONE:
 	if (show_named_timezones) {
-	    gtk_tree_store_insert_with_values (store, NULL, NULL, -1,
+	    gtk_list_store_insert_with_values (store, NULL, -1,
 					       LOC_GWEATHER_LOCATION_ENTRY_COL_LOCATION, loc,
-					       LOC_GWEATHER_LOCATION_ENTRY_COL_DISPLAY_NAME, loc->local_name,
-					       LOC_GWEATHER_LOCATION_ENTRY_COL_LOCAL_COMPARE_NAME, loc->local_sort_name,
-					       LOC_GWEATHER_LOCATION_ENTRY_COL_ENGLISH_COMPARE_NAME, loc->english_sort_name,
+					       LOC_GWEATHER_LOCATION_ENTRY_COL_DISPLAY_NAME, gweather_location_get_name (loc),
+					       LOC_GWEATHER_LOCATION_ENTRY_COL_LOCAL_SORT_NAME, gweather_location_get_sort_name (loc),
+					       LOC_GWEATHER_LOCATION_ENTRY_COL_LOCAL_COMPARE_NAME, gweather_location_get_sort_name (loc),
+					       LOC_GWEATHER_LOCATION_ENTRY_COL_ENGLISH_COMPARE_NAME, gweather_location_get_english_sort_name (loc),
 					       -1);
 	}
 	break;
@@ -699,12 +726,16 @@ match_selected (GtkEntryCompletion *completion,
 		GtkTreeIter        *iter,
 		gpointer            entry)
 {
-    if (model != ((GWeatherLocationEntry *)entry)->priv->model) {
+    GWeatherLocationEntryPrivate *priv;
+
+    priv = ((GWeatherLocationEntry *)entry)->priv;
+
+    if (model != priv->model) {
 	GeocodePlace *place;
 	char *display_name;
 	GeocodeLocation *loc;
 	GWeatherLocation *location;
-	GWeatherLocation *scope;
+	GWeatherLocation *scope = NULL;
 	const char* country_code;
 
 	gtk_tree_model_get (model, iter,
@@ -713,17 +744,17 @@ match_selected (GtkEntryCompletion *completion,
 			    -1);
 
 	country_code = geocode_place_get_country_code (place);
-	if (country_code != NULL)
-	    scope = gweather_location_find_by_country_code (gweather_location_get_world (), country_code);
-	else
-	    scope = gweather_location_get_world ();
+	if (country_code != NULL && gweather_location_get_level (priv->top) == GWEATHER_LOCATION_WORLD)
+	    scope = gweather_location_find_by_country_code (priv->top, country_code);
+	if (!scope)
+	    scope = priv->top;
 
 	loc = geocode_place_get_location (place);
 	location = gweather_location_find_nearest_city (scope, geocode_location_get_latitude (loc), geocode_location_get_longitude (loc));
 
-	location = create_new_detached_location(location, display_name, TRUE,
-						geocode_location_get_latitude (loc) * M_PI / 180.0,
-						geocode_location_get_longitude (loc) * M_PI / 180.0);
+	location = _gweather_location_new_detached (location, display_name, TRUE,
+						     geocode_location_get_latitude (loc) * M_PI / 180.0,
+						     geocode_location_get_longitude (loc) * M_PI / 180.0);
 
 	set_location_internal (entry, model, NULL, location);
 
@@ -755,9 +786,10 @@ fill_store (gpointer data, gpointer user_data)
     normalized = g_utf8_normalize (display_name, -1, G_NORMALIZE_ALL);
     compare_name = g_utf8_casefold (normalized, -1);
 
-    gtk_tree_store_insert_with_values (user_data, NULL, NULL, -1,
+    gtk_list_store_insert_with_values (user_data, NULL, -1,
 				       PLACE_GWEATHER_LOCATION_ENTRY_COL_PLACE, place,
 				       PLACE_GWEATHER_LOCATION_ENTRY_COL_DISPLAY_NAME, display_name,
+				       PLACE_GWEATHER_LOCATION_ENTRY_COL_LOCAL_SORT_NAME, compare_name,
 				       PLACE_GWEATHER_LOCATION_ENTRY_COL_LOCAL_COMPARE_NAME, compare_name,
 				       -1);
 
@@ -773,7 +805,7 @@ _got_places (GObject      *source_object,
     GList *places;
     GWeatherLocationEntry *self = user_data;
     GError *error = NULL;
-    GtkTreeStore *store = NULL;
+    GtkListStore *store = NULL;
     GtkEntryCompletion *completion;
 
     places = geocode_forward_search_finish (GEOCODE_FORWARD (source_object), result, &error);
@@ -792,7 +824,9 @@ _got_places (GObject      *source_object,
     }
 
     completion = gtk_entry_get_completion (user_data);
-    store = gtk_tree_store_new (4, G_TYPE_STRING, GEOCODE_TYPE_PLACE, G_TYPE_STRING, G_TYPE_STRING);
+    store = gtk_list_store_new (5, G_TYPE_STRING, GEOCODE_TYPE_PLACE, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
+    gtk_tree_sortable_set_default_sort_func (GTK_TREE_SORTABLE (store),
+                                             tree_compare_local_name, NULL, NULL);
     g_list_foreach (places, fill_store, store);
     g_list_free (places);
     gtk_entry_completion_set_match_func (completion, new_matcher, NULL, NULL);
@@ -842,45 +876,3 @@ gweather_location_entry_new (GWeatherLocation *top)
 			 "top", top,
 			 NULL);
 }
-
-static GWeatherLocation *
-create_new_detached_location (GWeatherLocation *nearest_station,
-                              const char       *name,
-                              gboolean          latlon_valid,
-                              gdouble           latitude,
-                              gdouble           longitude)
-{
-    GWeatherLocation *self;
-    char *normalized;
-
-    self = g_slice_new0 (GWeatherLocation);
-    self->ref_count = 1;
-    self->level = GWEATHER_LOCATION_DETACHED;
-    self->english_name = g_strdup (name);
-    self->local_name = g_strdup (name);
-
-    normalized = g_utf8_normalize (name, -1, G_NORMALIZE_ALL);
-    self->english_sort_name = g_utf8_casefold (normalized, -1);
-    self->local_sort_name = g_strdup (self->english_sort_name);
-    g_free (normalized);
-
-    self->parent = nearest_station;
-    self->children = NULL;
-
-    if (nearest_station)
-	self->station_code = g_strdup (nearest_station->station_code);
-
-    g_assert (nearest_station || latlon_valid);
-
-    if (latlon_valid) {
-	self->latlon_valid = TRUE;
-	self->latitude = latitude;
-	self->longitude = longitude;
-    } else {
-	self->latlon_valid = nearest_station->latlon_valid;
-	self->latitude = nearest_station->latitude;
-	self->longitude = nearest_station->longitude;
-    }
-
-    return self;
-}
